@class BmLocation ;
@class Browfile ;


// I make the following #defines instead of int constants so I can add them in 
// the enum NodeType below, and also so I don't have to declare stupid extern const
// in other files.

// Ordered sequence, used to test per browserv against threshold maximumNodeOrderAllowedAtRoot
#define BookdogNodeTypeOrderMask	0x00ff // mask
#define BookdogUndefinedOrder		0x0000  
#define BookdogRootOrder			0x0001 
#define BookdogBarOrder				0x0002
#define BookdogMenuOrder			0x0003
#define BookdogUnfiledOrder			0x0004
#define BookdogSharedOrder			0x0005
#define Bookdog1PasswordOrder		0x0006  
#define BookdogFolderOrder			0x0010 
#define BookdogBookmarkOrder		0x0020
#define BookdogSeparatorOrder		0x0030

// Other attributes of NodeType
#define BookdogNodeTypeCanHaveURLMask					0x0100
#define BookdogNodeTypeIsContainerMask					0x0200
#define BookdogNodeTypeIsMoveableMask					0x0400
#define BookdogNodeTypeIsEditableMask					0x0800
#define BookdogNodeTypeDisplayInNonhierarchicalViewMask	0x1000

enum NodeType
{                      //  order + [any other attributes applicable to this NodeType]                      
	BookdogUndefined	= BookdogUndefinedOrder,                                // 0x0000  
	BookdogRoot			= BookdogRootOrder + BookdogNodeTypeIsContainerMask,    // 0x0201 
	BookdogBar			= BookdogBarOrder + BookdogNodeTypeIsContainerMask,     // 0x0202
	BookdogMenu			= BookdogMenuOrder + BookdogNodeTypeIsContainerMask,    // 0x0203
	BookdogUnfiled		= BookdogUnfiledOrder + BookdogNodeTypeIsContainerMask, // 0x0204
	BookdogShared		= BookdogSharedOrder + BookdogNodeTypeIsContainerMask,  // 0x0205
	Bookdog1Password	= Bookdog1PasswordOrder + BookdogNodeTypeCanHaveURLMask + BookdogNodeTypeDisplayInNonhierarchicalViewMask,  // 0x1106  
    BookdogFolder		= BookdogFolderOrder + BookdogNodeTypeIsContainerMask + BookdogNodeTypeIsMoveableMask + BookdogNodeTypeIsEditableMask,  // 0x0e10 
    BookdogBookmark		= BookdogBookmarkOrder + BookdogNodeTypeCanHaveURLMask + BookdogNodeTypeIsMoveableMask + BookdogNodeTypeIsEditableMask + BookdogNodeTypeDisplayInNonhierarchicalViewMask,     // 0x1d20
    BookdogSeparator	= BookdogSeparatorOrder + BookdogNodeTypeIsMoveableMask + BookdogNodeTypeIsEditableMask  // 0x0c30
	
} ;
// See related int const masks in .m file

@interface NSArray (ArrayOfBmItems) 

- (NSArray*)arrayOfBmItemsWithChildrenOfPermanentFixturesAddedToTop ;

@end

@interface BmItem : NSObject
{
	NSMutableDictionary* _dic ;
	id _bmDoc ;	// Weak reference, not retained, to avoid retain cycles
}
// Creating
	- (BmItem*)initWithNodeType:(enum NodeType)nodeType ;
	+ (BmItem*)bmItemNodeType:(enum NodeType)nodeType ;
	+ (BmItem*)bmItem ;
	- (BmItem*)shallowCopyWithoutCopyingBmDocIvar ;
// Low-level accessors (not recommended, for legacy code)
	- (void)setObject:(id)object forKey:(NSString*)key ;
	- (id)objectForKey:(NSString*)key ;
	- (void)removeObjectForKey:(NSString*)key ;
	- (void)addEntriesFromDictionary:(NSDictionary*)otherDic ;
	


// Property Getters
	// Objects (may return nil except as indicated)
	- (NSNumber*)active ; // bool, Opera, rare but seen in both folders and bookmarks
	- (NSDate*)adddate ; // Opera, Google
	- (NSNumber*)addDateFixedNumberFormat ; // For stupid Bookit, use in formattedTimeIntervalWithDate
	- (NSNumber*)alloweddupe ;
	- (NSString*)archivedictionary ;  // OmniWeb RSS, folders only
	- (NSNumber*)autotab ; // Bool, Camino & Safari, folders only
	- (id)bmDoc ; // This is actually calculated for all except the greatest ancestor which should have had its key set with setBmDoc:.  It may return nil if self or one of its ancestors has been removed and is an orphan.
	- (NSNumber*)caminostatus ; // Camino.  A "number" in the plist, but I've only seen 0 or 1
	- (NSString*)checkfrequency ; // string, OmniWeb.  'never', 'daily', more???
	- (NSNumber*)code ;  // Verify
	- (NSString*)comments ;
	- (NSString*)commentsAsRead ;
	- (NSString*)delegateclass ; // OmniWeb RSS, folders only
	- (NSNumber*)disposition ; // Verify
	- (NSNumber*)dockmenu ; // bool, Camino, folders only
	- (NSNumber*)expanded ; // Opera, folders only
	- (NSNumber*)falsealarm ;
	- (NSString*)favicon ;  
	- (NSString*)faviconurl ;  // Camino, Opera
	- (NSString*)feedurl ;
	- (NSString*)feedurlnormalized ;
	- (id)innerstupidurl ; // Safari, only used in rare cases where the inner URL is different than the outer URL
	- (NSNumber*)insidebar ; // bool, Firefox, rare!
	- (NSNumber*)isshared ; // bool, Delicious, bookmarks only. isShared=nil if 'shared'="no", else isShared=oYES 
	- (NSString*)lastCharSet ;
	- (NSString*)lastcheckedtime ; // OmniWeb, both bookmarks and folders.  How do you check a folder??
	- (NSDate*)lastmodifieddate ; // Firefox, both bookmarks and folders.  Note: OmniWeb has a 'last-modified' attribute inside of its "validator" tag for RSS items, but I keep these separate.
	- (NSNumber*)lastModifiedDateFixedNumberFormat ; // For stupid Bookit, use in formattedTimeIntervalWithDate
	- (NSDate*)lastvisiteddate ;  // Firefox, OmniWeb; bookamrks only
	- (NSNumber*)lastVisitedDateFixedNumberFormat ; // For stupid Bookit, use in formattedTimeIntervalWithDate
	- (NSNumber*)master ; // bool, Bookdog, used during -migrateMigration to consolidate folders
	- (NSString*)name ;  // will not return nil
	- (NSString*)nameMayBeNil ;
	- (NSString*)nsErrorDomain ; // used during Verify 
	- (NSNumber*)noH3Tags ; // for fetchback.  Ordinarily, in OmniWeb and Firefox, folders have <H3> tags.  However, in OmniWeb, an item which does not have a href= attribute is always a folder
		// whether or not it has H3 tags.  You can create such a funny "fookmark" in OmniWeb Manage Bookmarks by
		// (1) create a new bookmark (2) delete all characters in its UFL and (3) watch closely.  You'll see that its
		// icon suddenly turns from a bookmark into a folder.
	- (NSString*)omniwebforms ;  // OmniWeb, bookmarks only, looks like maybe encrypted values for filling out a form?
  	- (NSString*)omniwebstatus ;  // OmniWeb, bookmarks and folders.  Folders are rare.
	- (NSNumber*)operaid ; // Nonpersistent (useless) ID used by Opera
	- (NSNumber*)operauniqueid ; // Persistent ID used by Opera, but only for their default bookmarks (so it is useless)
	- (NSNumber*)operaIDCreateIfNone:(BOOL)yn ; // will not return nil if yn = YES
	- (NSNumber*)operalinkbarfolder ; // bool, Opera, folders only
	- (NSNumber*)operalinkbarstop ; // bool, Opera, folders only
	- (NSNumber*)operaorder ;  // int, Opera, bookmarks and folders
	- (NSNumber*)operapanelposition ; // int, Opera.  Only seen in a few bookmarks but suspect could be in folders too
	- (NSNumber*)operapersonalbarposition ; // int, Opera, bookmarks and folders
	- (NSNumber*)operaselected ; // bool, Opera.  Only seen in 2 bookmarks but suspect could be in folders too.
	- (NSNumber*)operasmallscreen ; // bool, Opera.  Can't find this in any of my Opera fetchbacks!  Must be very rare
	- (NSNumber*)operatrashfolder ; // bool, Opera
	- (id)otherID ; // Returns the		for an Opera item
	- (BmItem*)parent ; // all bmItems except root should have their parent set upon creation, so this should not return nil except if rootoot
	- (NSString*)persistentidentifier ; 
	- (NSString*)persistentIdentifierCreateIfNoneTryHard:(NSNumber*)tryHard ;
	- (NSString*)postdata ; // Firefox extension "Habari Xenu", bookmarks only
	- (NSString*)reason ;  // Verify
	- (NSString*)shortcut ;  // ("Short Name") ma.gnolia, OmniWeb and Firefox("Keyword"), Camino ("Shortcut"), Opera ("Nickname")
	- (NSNumber*)sorted ;
	- (NSSet*)tags ;
	- (NSString*)url ;  // will not return nil
	- (NSString*)urlMayBeNil ;
	- (NSString*)urlnormalized ;
	- (NSString*)urlprior ;  // Verify
	- (NSString*)urlProposed ;  // Verify
	- (NSNumber*)uuidInCommentsOnly ;  // used for Opera, so Fetchback will not fail
	- (NSString*)validator ; // OmniWeb, in RSS, both feeds (folders) and items (bookmarks)
	- (NSNumber*)visitcount ; // OmniWeb, Camino, bookmarks only
	- (NSString*)webBookmarkIdentifier ; // Safari.  Only seen in folders so far, but because of the name I check for it in bookmarks too.
	- (NSString*)wordcount ; // OmniWeb, both bookmarks (common) and folders (rarely, only as RSS feeds)

	// Non-objects
	- (enum NodeType)nodeType ;
	- (BOOL)nodeTypeCanHaveURL ;
	- (BOOL)nodeTypeIsContainer ;
	- (BOOL)nodeTypeIsMoveable ;
	- (BOOL)nodeTypeIsEditable ;
	- (BOOL)isInBar ;
	- (BOOL)isInMenu ;

// Shallow (non-recursive) Calculations
- (NSNumber*)sortedKeyIfForced ;
- (BmItem*)childAtIndex:(int)index ;
- (NSMutableArray*)children ;
- (BmItem*)childNamed:(NSString*)name ;
- (BOOL)isSiblingOf:(BmItem*)otherItem ;
- (enum FixDisposition)fixDisposition ;
- (int)index ;
- (BmItem*)firstChildOfNodeType:(enum NodeType)nodeType ;
- (BOOL)isRoot ;
- (BOOL)isSorted_Legacy ;
- (BmItem*)previousSibling ;
- (BmItem*)nextSibling ;
- (BOOL)canAdoptChild:(BmItem*)item ;
#ifndef NO_GUI
- (BmLocation*)location ;
- (BmLocation*)locationUp1 ;
- (BmLocation*)locationDown1 ;
- (BmItem*)nextManualFixBookmark ;
#endif
- (int)numberOfChildren ;
- (NSArray*)subfolders ;
- (int)numberOfSubfolders ;
- (NSString*)listOfNamesOfChildren ;
- (int)indexOfBmItemWithSameNormalizedURLInArray:(NSArray*)array ;
- (BmItem*)bmItemWithSameNormalizedURLInArray:(NSArray*)array ;
- (BmItem*)bmItemWithSameNormalizedURLInSet:(NSSet*)set ;

// Deep (recursive) Calculations
- (NSString*)parentName ;
- (NSString*)browservName ;
- (BOOL)hasIncestWithProposedChild:(BmItem*)child ;
- (BOOL)hasIncestWithAnyOfProposedChildren:(NSArray*)children ;
- (NSArray*)lineageNamesMinimumNodeType:(enum NodeType)minNodeType
							includeSelf:(BOOL)includeSelf ; 
		   // Returns an array of names: ([self name], [parent name], ... ).
		   // Returns an empty array, not nil, if none are found
		   // This is required in parentingAction::::::::
- (NSString*)lineageLine ;  // lineage on one line: ([root name], ... , [parent name]).
	// If no lineage, returns an empty string.
- (BmItem*)collection ; // returns the collection-level object which contains this item, or self if self is a collection, or root if self is a bookmark (non-container) whose parent is root.  (This latter case is allowed in Firefox.)
- (struct FolderStatistics)addFolderStatistics ;
#ifndef NO_GUI
- (NSString*)displayStatistics ;
#endif
- (void)brokenStats:(struct BrokenBookmarksStatistics*)stats ;
//- (BOOL)containsItemActually:(BmItem*)item ;
- (NSMutableDictionary*)reformatToSameFormatSameParents:(BOOL)sameParents sameIDs:(BOOL)sameIDs ;
- (NSMutableDictionary*)reformatToSameFormatSameParentsSameIDs ;
- (NSMutableDictionary*)reformatToSameFormatNewParentsSameIDs ;
- (NSMutableDictionary*)reformatToSameFormatNewParentsNewIDs ;
- (NSString*)pathIndentedIncludingSelf:(BOOL)includeSelf ;

// Comparing
- (BOOL)marksSameSiteAs:(BmItem*)otherBookmark ;
- (NSComparisonResult)compareNodes:(BmItem *)otherDic ;
- (NSComparisonResult)compareCodes:(BmItem *)otherDic ;
- (BOOL)hasMoreInfoThan:(BmItem*)otherBookmark ;

// Shallow (non-recursive) Non-Mutating Actions

// looks like there are none anymore

// Deep (recursive) Non-Mutating Actions
- (void)flattenAndSortByNodeTypeImmoveableContainers:(NSMutableArray*)permanents
											 folders:(NSMutableArray*)folders
										   bookmarks:(NSMutableArray*)bookmarks
										  separators:(NSMutableArray*)separators
											 addTags:(id)tags ;
	// Pass in initialized NSMutableArrays; objects found will be added
	// If not interested in one or more of NodeTypes, pass nil as its arg
	// If tags responds to -containsObject: (is an NSAarray or NSSet),
	//      will add objects from array to the tags of each returned bookmark
	// If tags is any other non-nil value, will add to the tags of each returned
	//      bookmark the names of all non-immoveable parents in each item's lineage
	// If tags is nil, will not add any tags.
- (NSArray*)extractBookmarksOnly ;
	// Sends flattenAndSortByNodeTypeImmoveableContainers:folders:bookmarks:separators:addParentNamesAsTags:
	// with other arguments set to nil.
- (void)sortTree ;
- (void)updateLegacyFalseAlarms ;
- (BmItem*)deepOrphanedCopySameIDs:(BOOL)sameIDs ; // (returns a copy, not autoreleased)

// Property Setters
	// Objects (OK to send nil to any of these methods; they will remove the property key/value pair)
	- (void)setActive:(NSNumber*)yn ;
	- (void)setAdddate:(NSDate*)_ad ;
	- (void)setAddDateFixedNumberFormat:(NSNumber*)format ; // For stupid Bookit
	- (void)setAlloweddupe:(NSNumber*)yn ;
	- (void)setArchivedictionary:(NSString*)x ;
	- (void)setAutotab:(NSNumber*)yn ;
	- (void)setBmDoc_Weakly:(id)bmDoc_ ; // Only used for root bmItem; others derive it by looking "up",
										 // using -parent recursively until the root is reached.
										 // Weak reference; does not retain, to avoid retain cycle
	- (void)setCaminostatus:(NSNumber*)_cs ;
	- (void)setCheckfrequency:(NSString*)x ;
	- (void)setChildren:(NSMutableArray*)children_ ;
	- (void)setCode:(NSNumber*)code_ ;
	- (void)setComments:(NSString*)comments_ ;
	- (void)setCommentsAsRead:(NSString*)x ;
	- (void)setDelegateclass:(NSString*)x ;
	- (void)setDisposition:(NSNumber*)disposition_ ;
	- (void)setDockmenu:(NSNumber*)yn ;
	- (void)setExpanded:(NSNumber*)yn ;
	- (void)setFalsealarm:(NSNumber*)yn ;
	- (void)setFavicon:(NSString*)fi_ ;
	- (void)setFaviconurl:(NSString*)x ;  
	- (void)setFeedurl:(NSString*)fu_ ;
	- (void)setInsidebar:(NSNumber*)yn ;
	- (void)setInnerstupidurl:(id)isu_ ;
	- (void)setIsshared:(NSNumber*)yn ;
	- (void)setTags:(NSSet*)tags ;
	- (void)appendTags:(id)additionalTags ; // additionalTags may be NSArray* or NSSet*
	- (void)setLastCharSet:(NSString*)lcs_ ;
	- (void)setLastcheckedtime:(NSString*)x ;
	- (void)setLastmodifieddate:(NSDate*)_lastModified ;
	- (void)setLastModifiedDateFixedNumberFormat:(NSNumber*)format ; // For stupid Bookit
	- (void)setLastvisiteddate:(NSDate*)_lastVisitedDate ;
	- (void)setLastVisitedDateFixedNumberFormat:(NSNumber*)format ; // For stupid Bookit
	- (void)setMaster:(NSNumber*)yn ;
	// Argument-less wrappers for setMaster
		-(void)setMasterYES ;
		-(void)setMasterNO ;
	- (void)setName:(NSString*)name_ ;
	- (void)setNoH3Tags:(NSNumber*)x ;
	- (void)setNSErrorDomain:(NSString*)nsed_ ;
	- (void)setOmniwebforms:(NSString*)x ;
	- (void)setOmniwebstatus:(NSString*)x ;
	- (void)setOperaorder:(NSNumber*)x ;
	- (void)setOperaid:(NSNumber*)x ;
	- (void)setOperauniqueid:(NSString*)x ;
	- (void)setOperalinkbarfolder:(NSNumber*)x ;
	- (void)setOperalinkbarstop:(NSNumber*)x ;
	- (void)setOperapanelposition:(NSNumber*)x ;
	- (void)setOperapersonalbarposition:(NSNumber*)x ;
	- (void)setOperaselected:(NSNumber*)yn ;
	- (void)setOperasmallscreen:(NSNumber*)yn ;
	- (void)setOperatrashfolder:(NSNumber*)x ;
	- (void)setOperauniqueid:(NSString*)x ;
	- (void)setParent:(BmItem*)parent ;
	- (void)setPersistentidentifier:(NSString*)uuid_ ;
	- (void)setPostdata:(NSString*)fu_ ;
	- (void)setReason:(NSString*)_reason ;
	- (void)setShortcut:(NSString*)_shortcut ;
	- (void)setSorted:(NSNumber*)yn ;
	- (void)setUrl:(NSString*)url_ ;
	- (void)setUrlprior:(NSString*)urlprior_ ;
	- (void)setUrlProposed:(NSString*)urlProposed_ ;
	- (void)setUuidInCommentsOnly:(NSNumber*)x ;
	- (void)setValidator:(NSString*)x ;
	- (void)setVisitcount:(NSNumber*)visitcount ;
	- (void)setWebBookmarkIdentifier:(NSString*)webBookmarkIdentifier_ ;
	- (void)setWordcount:(NSString*)x ;

	// Non-objects
	- (void)setNodeType:(enum NodeType)nodeType_ ;

// Shallow (non-recursive) Mutating Actions
// These methods do NOT:
//   set up undo, tellTreeModified, incrementMyChangeCount, update windows, show items, or order windows.
- (BOOL)appendToBookdogParent:(BmItem*)newParent
				  onlyAtIndex:(unsigned int)newIndex ;
	// If newIndex is the next available index of newParent,
	//    adds self to children of newParent, sets self's KEY_PARENT, and returns YES
	// If cannot set at self at newIndex because not enough siblings,
	//    does nothing and returns NO ;
	// If newParent is nil, does nothing and returns NO ;
	// Does not remove from previous parent!  Use moveToBookdogParent: for that
- (void)setToBookdogParent:(BmItem*)newParent atIndex:(unsigned int)newIndex ;
	// Adds self to children of newParent, and sets self's KEY_PARENT
	// To add as last child, pass newIndex=NSNotFound
	// If cannot set at newIndex because there are not that many siblings,
	//    adds as last sibling instead
	// To make orphan, pass newParent = nil ;
	// Does not remove from previous parent!  Use moveToBookdogParent: for that
- (void)setToBookdogParent:(id)newParent ;
	// Sends -setToBookdogParent atIndex:NSNotFound
	// Does not remove from previous parent!  Use moveToBookdogParent: forthat
- (void)moveToBookdogParent:(BmItem*)newParent atIndex:(unsigned int)newIndex ;
	// Sends setToBookdogParent but also remove self from previous parent,
	// To add as last sibling, pass newIndex = NSNotFound
- (void)moveToBookdogParent:(BmItem*)newParent ;
	// Sends -moveToBookdogParent atIndex:NSNotFound
- (void)remove ;
- (int)removeAndRemoveChildlessAncestors ;
	// Note that removing an item may make its ancestors childless
	// This method removes childless ancestors, but only if they are
	// of type BookdogFolder.  (Do not remove permanent fixtures).
	// Returns the number of ancestors removed.
- (void)setSortedAndAllowedDupeConvertingFromLegacy ;
- (void)removeLegacySortedUnsortedAndAllowedDupeNameTags ;
- (void)removeIfIsSeparator ;
- (void)removeIfIsEmptyFolder ;
- (void)consolidateFolders ;
- (void)moveFeedURLToURL ;
- (void)removeIfSeparator ;
- (void)regeneratePersistentidentifier ;
	// Will return a new persistentidentifier for item, or the existing
	// persistentidentifier.  Which you get follows rules given in header 
	// documentation for -[Store freshPersistentidentifierForItem:tryHard:]. 
- (void)replaceDelimiterSpoofsInTagsWithUnderscoreForBrowfile:(Browfile*)browfile ;

// Deep (recursive) Mutating Actions
- (void)ifPersonalBarPositionMoveToBar:(BmItem*)bar ;
- (void)recursivelyPerformSelector:(SEL)selector ;
- (void)recursivelyDepthFirstPerformSelector:(SEL)selector ;
- (void)recursivelyPerformSelector:(SEL)selector withObject:(id)object ;
- (void)recursivelyPerformSelector:(SEL)selector withObject:(id)object1 withObject:(id)object2 ;
- (void)recursivelyPerformSelector:(SEL)selector withObject:(id)object toDepth:(int)maxDepth currentDepth:(int)currentDepth ;
- (void)removeAllSeparators ;
- (void)createFoldersFromTagsMultiple:(BOOL)doMultiple ;
	// assumes that self is flat; i.e. only children, no grandchildren
	// If doMultiple == YES, creates a folder for each tag
	// If not, arbitrarily picks only one

// Changing View
- (void)expandInOutline:(NSOutlineView*)outlineView ;

// Debugging
- (NSString*)description ;
- (NSString*)descriptionDeep ;
- (void)debugAddressAndDescriptionToLog ;

@end